// ledarray_master.c
// for NerdKits with ATmega168
// hevans@nerdkits.com

#define F_CPU 14745600

#include <stdio.h>

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>

#include "../libnerdkits/delay.h"
#include "../libnerdkits/uart.h"

#include "font.h"

// PIN DEFINITIONS:
//
// PB5 - SCK
// PB3 - MOSI (Master Out Slave In)
// PB1, PC0-2 SS pins for slave panels

#define ROWS 5
#define COLS_PER_ARRAY 20
#define NUM_ARRAYS 4
#define COLS (NUM_ARRAYS * COLS_PER_ARRAY)

//keeps the entire array. NUM_ARRAYS * COLS_PER_ARRAY
uint8_t data[COLS];

//sets all columns in the array to 0
void blank_data(){
  uint16_t i = 0;
  for(i=0;i<COLS;i++){
    data[i] = 0;
  }
}

void data_left_shift(){
  uint16_t i = 0;
  for(i=0;i<COLS-1;i++){
    data[i] = data[i+1];
  }
  data[COLS-1] = 0;
}

void all_on(){
  uint16_t i;
  for(i=0;i<COLS;i++){
    data[i] = 0x3f;
  }
}

inline uint8_t data_get(uint8_t i, uint8_t j) {
  if(i < ROWS && j < COLS) {
    if((data[j] & (1<<i)) != 0) {
      return 1;
    } else {
      return 0;
    }
  } else {
    return 0;
  }
}

inline void data_set(uint8_t i, uint8_t j, uint8_t onoff) {
  if(i < ROWS && j < COLS) {
    if(onoff) {
      data[j] |= (1<<i);
    } else {
      data[j] &= ~(1<<i);
    }
  }
}
void font_get(char match, char *buf) {
  // copies the character "match" into the buffer
  uint8_t i;
  PGM_P p;
  
  for(i=0; i<FONT_SIZE; i++) {
    memcpy_P(&p, &font[i], sizeof(PGM_P));
    
    if(memcmp_P(&match, p,1)==0) {
      memcpy_P(buf, p, 7);
      return;
    }
  }
  
  // NO MATCH?
  font_get('?', buf);
}

uint8_t font_width(char c) {
  char buf[7];
  buf[1] = 0;
  
  font_get(c, buf);
  
  return buf[1];
}

void font_display(char c, uint8_t offset) {
  char buf[7]; 
  font_get(c, buf);
  
  uint8_t width = buf[1];
  uint8_t i, j;
  for(i=0; i<ROWS; i++) {
    for(j=0; j<width; j++) {
      if((offset + j) < COLS) {
        if( (buf[2+j] & (1<<i)) != 0) {
          data_set(i,offset + j,1);
        } else {
          data_set(i,offset + j,0);
        }
      }
    }
  }
  
  // blank the next column to the right
  for(i=0; i<ROWS; i++) {
    data_set(i, offset+width, 0);
  }
}


void do_scrolling_display() {
  blank_data();
  int16_t offset = 0, next_offset = 0;
  uint16_t is_started = 0;
  char x=' ';
  
  while(1) {
    //we are started, shift the array, and display the next character
    if(is_started) {
      delay_ms(60);
      data_left_shift(); 
      if(next_offset > 0) {
        offset -= 1;
        next_offset -= 1;
      }
      font_display(x, offset); 
    } else { //we are not started, just shift the next offset
      offset = COLS-1;
    }

    // if we can now accept a new character, tell the computer
    if(next_offset == COLS)
      uart_write('n');
    
    while(uart_char_is_waiting()) {
      if(is_started)
        offset = next_offset;
      
      x = uart_read(); 
      if(x=='a') {
        blank_data();
        return;
      }
      font_display(x, offset);
      next_offset = offset + font_width(x)+1;
      is_started = 1;
       
      // if we can now accept a new character, tell the computer
      if(next_offset <= COLS)
        uart_write('n');
      
    }
    
  }
}

void master_init(){

  //set MOSI,SCK,SS as output
  DDRB |= (1<<PB3) | (1<<PB5) | (1<<PB2);
  //initiate the SPI module in master mode, data rate clk/16
  SPCR |= (1<<SPE) | (1<<MSTR) | (1<<SPR0);

  //set the SS pins we are using as outputs
  DDRC |= (1<<PC0) | (1<<PC1) | (1<<PC2) | (1<<PC3) | (1<<PC4) | (1<<PC5);
  DDRB |= (1<<PB1);
  //set the SS pins logic high (slave not active)
  PORTC |= (1<<PC0) | (1<<PC1) | (1<<PC2) | (1<<PC3) | (1<<PC4) | (1<<PC5);
  PORTB |= (1<<PB1);

  //set up timer, when this fires all the slave panels are updated
  //timer set to clk/1024 - aprox 56Hz 
  TCCR0B |= (1<<CS02) | (1<<CS00);
  TIMSK0 |= (1<<TOIE0);
 
}


//pull the slave line now corresponding to the ith panel (active)
inline void activate_array_ss(uint8_t i){
  if(i == 0){
    PORTB &= ~(1<<PB1);
  } else if (i < 7){
    PORTC &= ~(1<<(PC0+i-1));
  }
  delay_us(5);
}

//pull the slave line high corresponding ot the ith panel (not active)
inline void deactivate_array_ss(uint8_t i){
   if(i == 0){
    PORTB |= (1<<PB1);
  } else if(i < 7){
    PORTC |= (1<<(PC0+i-1));
  } 
  delay_us(5);
}


//timer has overflowed, update all the panels
SIGNAL(SIG_OVERFLOW0) {
  uint16_t i = 0;
  uint16_t j = 0;

  //cycle over the arrays
  for(i=0;i<NUM_ARRAYS;i++){
  
    //pull SS low - activate slave
    activate_array_ss(i);

    //cycle over each byte to send to this array.
    for(j=0;j<COLS_PER_ARRAY;j++){

      //grab the next byte and send it
      SPDR = data[i*COLS_PER_ARRAY+j];

      //wait for transmition to be over
      while(!(SPSR & (1<<SPIF))){
        
      }
      //wait a little bit before sending the next byte
      delay_us(15);
    }
    //send done byte
    SPDR = 0xff;
    //wait for transmition to be over
    while(!(SPSR & (1<<SPIF))){
        
    }

    //pull SS high - deactivate slave
    deactivate_array_ss(i);

  }
}

int main() {
  //ledarray_init();
  master_init();
  // activate interrupts
  sei();

  // init serial port
  uart_init();
  FILE uart_stream = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);
  stdin = stdout = &uart_stream;
  
  // mode 1: test pattern
  //do_testpattern();

  do_scrolling_display();
  while(1) {
    //this loop never executes, scrolling display should loop forever in this mode of operation
    //everythign exciting happens in the interrupt handler, and in do_scrolling display
  }
  return 0;
}

